import trueskill, sys
import numpy as np
import pandas as pd
import networkx as nx
from sklearn.metrics import roc_curve, auc
from dateutil.relativedelta import relativedelta
from sklearn.linear_model import LogisticRegression

def draw_sigmoid(training, model, day):
   df = pd.DataFrame(data = np.arange(training["home_winner_score"].min() * 10, training["home_winner_score"].max() * 10, 0.01), columns = ("home_winner_score",))
   df["home_win"] = model.predict_proba(df[["home_winner_score",]])[:,1]
   df.to_csv(f"{day.year}_logistic.tsv", sep = "\t", index = False)

def _train_model(training, predictor):
   if predictor == "pagerank":
      training, ranks = _train_pagerank(training)
   elif predictor == "elo":
      training, ranks = _train_elo(training)
   elif predictor == "naive":
      training, ranks = _train_naive(training)
   training["home_winner_score"] = training["home_rank"] - training["away_rank"]
   model = LogisticRegression(penalty = "none")
   model.fit(training[["home_winner_score",]], training["home_win"])
   return model, ranks

def _train_pagerank(training):
   G = training.groupby(by = ["winner", "loser"]).size().reset_index().rename(columns = {0: "weight"})
   G = nx.from_pandas_edgelist(G, source = "loser", target = "winner", edge_attr = "weight", create_using = nx.DiGraph())
   ranks = nx.pagerank(G)
   training["home_rank"] = training["home_team"].map(ranks)
   training["away_rank"] = training["away_team"].map(ranks)
   return training, ranks

def _train_elo(training):
   env = trueskill.TrueSkill(draw_probability = 0)
   home_elos = []
   away_elos = []
   ranks = {}
   team_current_rating = {}
   for index, row in training.iterrows():
      if not row["home_team"] in team_current_rating:
         team_current_rating[row["home_team"]] = trueskill.Rating()
      if not row["away_team"] in team_current_rating:
         team_current_rating[row["away_team"]] = trueskill.Rating()
      home_elos.append(team_current_rating[row["home_team"]].mu)
      away_elos.append(team_current_rating[row["away_team"]].mu)
      team_current_rating[row["winner"]], team_current_rating[row["loser"]] = trueskill.rate_1vs1(team_current_rating[row["winner"]], team_current_rating[row["loser"]], env = env)
      ranks[row["home_team"]] = team_current_rating[row["home_team"]].mu
      ranks[row["away_team"]] = team_current_rating[row["away_team"]].mu
   training["home_rank"] = home_elos
   training["away_rank"] = away_elos
   return training, ranks

def _train_naive(training):
   wins = training.groupby(by = "winner").size()
   losses = training.groupby(by = "loser").size()
   ranks = wins.combine(losses, lambda x, y: x / (x + y), fill_value = 0).to_dict()
   training["home_rank"] = training["home_team"].map(ranks)
   training["away_rank"] = training["away_team"].map(ranks)
   return training, ranks

def _test_model(df, day, model, ranks):
   test = df[df["date"] == day].copy()
   test["home_rank"] = test["home_team"].map(ranks)
   test["away_rank"] = test["away_team"].map(ranks)
   test = test.dropna() #TODO: Dropping NAs from games with teams not in the training data
   if test.shape[0] > 0:
      test["home_winner_score"] = test["home_rank"] - test["away_rank"]
      test["home_winner_score"] = model.predict_proba(test[["home_winner_score",]])[:,1]
      return test

def prediction_scores(df, predictor):
   df["date"] = pd.to_datetime(df["date"], format = "%Y%m%d")
   df = df.sort_values(by = "date")
   start_prediction = df["date"].min() + relativedelta(years = 1)
   results = []
   home_advs = []
   for day in df[df["date"] >= start_prediction]["date"].drop_duplicates():
      training = df[(df["date"] < day) & (df["date"] >= (day - relativedelta(years = 1)))].copy()
      if training.shape[0] > 1:
         model, ranks = _train_model(training, predictor)
         if model is not None:
            home_advs.append(model.predict_proba(pd.DataFrame(data = [(0,)], columns = ("home_winner_score",)))[:,1][0])
            results.append(_test_model(df, day, model, ranks))
   if len(results) > 0:
      return pd.concat(results).dropna(), home_advs
   else:
      return pd.DataFrame(), []

def predict(df, predictor = "pagerank"):
   home_adv = []
   scores = []
   for index, league in df["league"].drop_duplicates().items():
      sys.stderr.write(f"{league}...\n")
      league_scores, league_home_adv = prediction_scores(df[df["league"] == league].copy(), predictor)
      home_adv.extend(league_home_adv)
      scores.append(league_scores)
   scores = pd.concat(scores)
   fpr, tpr, thresholds = roc_curve(scores["home_win"], scores["home_winner_score"])
   return auc(fpr, tpr), np.mean(home_adv)
